from __future__ import print_function

import os.path
import subprocess
import json
import copy
import threading
import time

CATTURA_PATH = "/var/cattura"
PROCESSING_AGENT = os.path.join(CATTURA_PATH, "processing-agent.jar")
PLUGIN_TYPES = ["analysis", "capturing", "encoding", "processor", "processing", "metadata", "publishing"]

GRAPH_JSON = "graph.json"

VIEWER_COMMAND = "less"

XUGGLER_DIRECTORY = os.path.join("/", "tmp", "xuggle")

class Runner(object):
    def __init__ (self, terminal_executor=None):
        self.processing_agent = PROCESSING_AGENT
        self.cattura_path = CATTURA_PATH
        self.plugin_types = PLUGIN_TYPES
        self.terminal_executor = terminal_executor

    def is_usable (self):
        return os.path.isfile(self.processing_agent)

    def __call__ (self, json_graph, target):
        dump_graph(json_graph, target)
        command = ["java", "-jar", self.processing_agent] + self.get_arguments("--graph", target, "--output", target)

        if self.terminal_executor:
            self.terminal_executor.execute(command)
        else:
            subprocess.call(command)

        while check_if_graph_is_running(target):
            time.sleep(1) # seconds

    def open (self):
        return ProcessingAgent(self.processing_agent, *self.get_arguments())

    def get_arguments (self, *args):
        return ["-p{}/plugins".format(self.cattura_path)] + \
               ["-p{}/plugins/{}".format(self.cattura_path, plugin_type) for plugin_type in self.plugin_types] + \
               list(args)

class ProcessingAgent(object):
    def __init__ (self, processing_agent, *args):
        self.outputHandler = print
        self.command = ["java", "-jar", processing_agent] + list(args)
        self.messageId = 0
        self.threaded = False
    
    def __enter__ (self):
        self.listeners = []
        return self

    def sendNotification (self, name, body, callback=None):
        if self.process is None:
            return

        messageId = str(self.messageId)

        if callback:
            def wrapped (notification):
                if notification['messageID'] == messageId:
                    callback(notification)
            self.listeners.append(wrapped)

        self.process.stdin.write(json.dumps({
            "notification": name,
            "body": body,
            "messageID": self.messageId
        }))
        self.process.stdin.write("\n")
        self.process.stdin.flush()
        self.messageId = self.messageId + 1

    def on (self, event):
        def register (func):
            if hasattr(event,'__call__'):
                def wrapped (notification):
                    if event(notification):
                        func(notification)
            else:
                def wrapped (notification):
                    if notification['notification'] == event:
                        func(notification)
            self.listeners.append(wrapped)
        return register

    def onOutput (self):
        def register (func):
            self.outputHandler = func
        return register
    
    def onMessage (self, notification):
        for listener in self.listeners:
            listener(notification)

    def run (self):
        while True:
            try:
                line = self.process.stdout.readline()
                try:
                    note = json.loads(line)
                    self.onMessage(note)
                except ValueError:
                    self.outputHandler(line.strip())
            except KeyboardInterrupt: break
            except ValueError: break

    def __exit__ (self, type, value, traceback):
        self.process = subprocess.Popen(
            self.command,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT
        )

        if self.threaded:
            threading.Thread(target=self.run).start()
        else:
            self.run()

    def terminate (self):
        if self.process is None:
            return

        self.process.stdout.close()
        self.process.terminate()
        self.process.wait()

def dump_graph (json_graph, target):
    if os.path.isdir(target):
        os.rmdir(target)
    
    os.makedirs(target)
    graph_path = os.path.join(target, GRAPH_JSON)
    with open(graph_path, 'wb') as f:
        json.dump(fix_working_directories(json_graph), f, indent=4, sort_keys=True, separators=(',', ': '))    

def fix_working_directories (graph_json):
    graph_json = copy.deepcopy(graph_json)
    for node in graph_json["nodes"]:
        del node["workingDir"]

    return graph_json

def check_if_graph_is_running (target):
    for line in subprocess.check_output(["ps", "-Af"]).strip().split("\n"):
        if PROCESSING_AGENT in line and target in line:
            return True
    return False

def test_xuggler_directory ():
    test_file_path = os.path.join(XUGGLER_DIRECTORY, "catgut")
    try:
        with open(test_file_path, "w") as test_file:
            test_file.write("")
        os.remove(test_file_path)
        return True
    except IOError as e:
        print(e)

    return False