import functools

def getConfigFieldText(config, path):
    parts = path.split('/')
    o = config
    for p in parts:
        o = o.get(p, None)
        if o is None:
            return '<font color="red">** %s not found **</font>' % (path,)
    return o

def element_is_of_type (element, element_type):
    element_class = element.get("@class", "")
    return element_class == element_type or element_class.endswith(".{}".format(element_type))

class KnownPlugin(object):
    def __init__ (self, plugin_ids, color, mouseover_text=None, config_generator=None):
        self.id = plugin_ids[0]
        self.also_known_as = plugin_ids[1:]
        self.simple_id = self.id.split(".")[-1]
        self.color = color
        self.mouseover_text_function = mouseover_text
        self.config_generator = config_generator

    def mouseover_text (self, config):
        if self.mouseover_text_function:
            gcf = functools.partial(getConfigFieldText, config)
            return self.mouseover_text_function(config, gcf)
        
        return display_map(config)

    def make_config (self, ask=lambda question: None):
        config = {
            "pluginID": self.id
        }

        if self.config_generator:
            generated_config = self.config_generator(self, ask)

            if generated_config:
                config.update(generated_config)

        return config

stitcherAttrs = {'sourceWidth'  : (u'VideoFrame', u'width'),
                 'sourceHeight' : (u'VideoFrame', u'height'),
                 'sourceID'     : (u'VideoFrame', u'sourceID'),

                 'canvasWidth'  : (u'Rectangle', u'width'),
                 'canvasHeight' : (u'Rectangle', u'height'),
                 'xOffset'      : (u'Rectangle', u'x'),
                 'yOffset'      : (u'Rectangle', u'y')
                 }

def display_video_format (element):
    return "\n".join([
        "resolution: {width}x{height}",
        "fps: {fps}",
        "source id: {sourceID}"
    ]).format(**element)

def display_audio_format (element):
    return "\n".join([
        "audio channels: {numberChannels}",
        "sample rate: {sampleRate}"
    ]).format(**element)

def display_map (map):
    return "\n".join(["{}: {}".format(key, value) for (key, value) in map.iteritems()])

def stitcher_mouseover (config, gcf):
    text = 'canvas resolution: %s' % (gcf('resolution'),)
    text += '\ntarget fps: %s' % (gcf('fps/value'),)
    text += '\n'
    text += '\nInputs:'

    for obj in config['inputs']['values']:
        if element_is_of_type(obj, "Group"):

            item = {}
            for attr, (klass, eAttr) in stitcherAttrs.items():
                for elem in obj['elements']:
                    if element_is_of_type(elem, klass):
                        if eAttr in elem:
                            item[attr] = elem[eAttr]
                
            text += '\n'
            text += '\nsource id: %s' % (item['sourceID'],)
            text += '\nsource resolution: %sx%s' % (item['sourceWidth'], item['sourceHeight'])
            text += '\ncanvas resolution: %sx%s' % (item['canvasWidth'], item['canvasHeight'])
            text += '\ncanvas x offset: %s' % (item['xOffset'],)
            text += '\ncanvas y offset: %s' % (item['yOffset'],)

    return text

def writer_mouseover (config, gcf):
    fields = []
    if 'audio' in config and 'video' in config:
        fields.append("[AUDIO/VIDEO]")
    elif 'audio' in config:
        fields.append("[AUDIO]")
    elif 'video' in config:
        fields.append("[VIDEO]")

    if "id" in config:
        fields.append("Output ID: {}".format(config['id']))

    if "name" in config:
        fields.append("Output Name: {}".format(config['name']))

    if 'audio' in config:
        fields.append("")
        fields.append(display_audio_format(config['audio']))

    if 'video' in config:
        fields.append("")
        fields.append(display_video_format(config['video']))
        fields.append("preset keyword: %s" % (gcf('preset_keyword'),))
    
    return "\n".join(fields)

def transmuxer_mouseover (config, gcf):
    return "\n".join([
        "id: {id}",
        "filename: {filename}.{extension}"
    ]).format(**config)

def resampler_mouseover (config, gcf):
    bounds = config.get("bounds", config.get("video", None))
    if not bounds:
        return ""
    
    return 'scale to: %sx%s' % (bounds['width'], bounds['height'])

def video_capture_mouseover (config, gcf):
    fields = []
    elements = config['device']['elements']
    for e in elements:
        if element_is_of_type(e, "VideoFrame"):
            fields.append(display_video_format(e))

        elif element_is_of_type(e, "Metadata"):
            fields.append('Driver: %s' % (e['map']['Driver'],))
            fields.append('Friendly Name: %s' % (e['map']['Friendly Name'],))
            fields.append('Device Path: %s' % (e['map']['Device Path'],))
    
    return "\n".join(fields)

def video_capture_generator (plugin, ask):
    source_id = ask("Video Source ID")
    if not source_id:
        raise Exception("Source ID is required")
    
    return {
        "device": {
            "@class": "com.cattura.share.processor.data_provider.elements.Group",
            "elements": [
                {
                    "@class": "com.cattura.share.processor.data_provider.elements.Metadata",
                    "map": {
                         "valid": True,
                         "name": source_id,
                         "Driver": "????",
                         "Device Path": source_id,
                         "Friendly Name": source_id
                    }
               },                
                {
                    "@class": "com.cattura.share.processor.data_provider.elements.VideoFrame",
                    "colorspace": [],
                    "fps": 0,
                    "height": -1,
                    "sourceID": source_id,
                    "width": -1
                }
            ]
        }
    }

def audio_capture_generator (plugin, ask):
    source_id = ask("Audio Source ID")
    if not source_id:
        raise Exception("Source ID is required")
    
    return {
        "device": {
            "@class": "com.cattura.share.processor.data_provider.elements.Group",
            "elements": [
                {
                        "@class": "com.cattura.share.processor.data_provider.elements.Metadata",
                        "map": {
                            "vendor": "????",
                            "name": source_id,
                            "usbInfo": "",
                            "version": "????",
                            "friendlyName": source_id
                        }
                },
                {
                        "@class": "com.cattura.share.processor.data_provider.elements.AudioFormat",
                        "sampleRate": 44100,
                        "numberChannels": 1,
                        "sampleSizeInBits": 50000,
                        "isSigned": False,
                        "isBigEndian": True
                }
            ]
        }
    }

def metadata_mouseover (config, gcf):
    element = config.get("metadata", config.get("element", config.get("asset", None)))
    if not element:
        return "[No Element]"
    
    fields = [
        "[{}]".format(element['@class'].split(".")[-1])
    ]
    if element_is_of_type(element, "Metadata"):
        map = element.get("map", element)
        fields.append(display_map(map))
    elif element_is_of_type(element, "Asset") or element_is_of_type("MediaAsset"):
        fields.append("path: {0}".format(element['path']))
    
    return "\n".join(fields)

def audio_capture_mouseover (config, gcf):
    fields = []
    elements = config['device']['elements']
    for e in elements:
        if element_is_of_type(e, "AudioFormat"):
            fields.append(display_audio_format(e))

        elif element_is_of_type(e, "Metadata"):
            fields.append('Friendly Name: %s' % (e['map']['friendlyName'],))
            fields.append('name: %s' % (e['map']['name'],))

    return "\n".join(fields)

def device_capture_mouseover (config, gcf):
    device = config['device']
    fields = []

    selected_audio_indices = device['audioFormats']['selected']
    selected_video_indices = device['videoFormats']['selected']

    if selected_audio_indices and selected_video_indices:
        fields.append("[AUDIO/VIDEO]")
    elif selected_audio_indices in config:
        fields.append("[AUDIO]")
    elif selected_video_indices in config:
        fields.append("[VIDEO]")

    fields.append("Device ID: {}".format(device['id']))

    if device['name']:
        fields.append("Device Name: {}".format(device['name']))

    if selected_audio_indices:
        selected_index  = selected_audio_indices[0]
        selected_format = device['audioFormats']['values'][selected_index]
        fields.append("\nSelected Audio Format Index: {}".format(selected_index))
        fields.append(display_audio_format(selected_format))

    if device['videoFormats']['selected']:
        selected_index  = device['videoFormats']['selected'][0]
        selected_format = device['videoFormats']['values'][selected_index]
        fields.append("\nSelected Video Format Index: {}".format(selected_index))
        fields.append(display_video_format(selected_format))

    return "\n".join(fields)

# File Reader Plugin
def file_reader_mouseover (config, gcf):
    return "\n".join([
        "path: {}",
        "mode: {}"
    ]).format(gcf('asset/path'), gcf('mode'))

def file_reader_generator (plugin, ask):
    mode = ask("File Reader Mode (freewheel or playback)")
    if not mode:
        mode = "playback"

    config = {
        "mode": mode
    }

    path = ask("File Path")
    if path:
        config['asset'] = {
            "@class": "com.cattura.share.processor.data_provider.elements.Asset",
            "path": path
        }

    return config

# Asset Plugin
def asset_generator (plugin, ask):
    path = ask("File Path")
    if not path:
        raise Exception("Path is required")
    
    return {
        "asset": {
            "@class": "com.cattura.share.processor.data_provider.elements.Asset",
            "path": path
        },
        "mode": mode
    }

def no_signal_mouseover (config, gcf):
    text = 'description: %s' % (gcf('no_signal_path/description'),)
    text += '\nfileAlias: %s' % (gcf('no_signal_path/fileAlias'),)
    text += '\npath: %s' % (gcf('no_signal_path/path'),)
    return text

def cws_mouseover (config, gcf):
    return 'resolution: %sx%s' % (gcf('boundary/width'), gcf('boundary/height'))

def save_frame_mouseover (config, gcf):
    text = 'preferredWidth: %s' % (gcf('preferredWidth'),)
    text += '\npreferredHeight: %s' % (gcf('preferredHeight'),)
    text += '\nfilename: %s' % (gcf('filename'),)
    text += '\nextension: %s' % (gcf('extension'),)
    return text

def text_generator_mouseover (config, gcf):
    return "\n".join([
        "id: {0}",
        "text: {1}",
        "font-size: {2}",
        "padding: {3}",
    ]).format(gcf('id'), gcf('text'), gcf('font-size'), gcf('padding'))

def fps_mouseover (config, gcf):
    fps = config['fps']
    try:
        fps = fps['value']
    except Error: pass

    return "fps: {0}".format(fps)

# ShotDetect
def shotdetect_mouseover (config, gcf):
    return "threshold: {0}".format(gcf('threshold'))

def shotdetect_generator (plugin, ask):
    try:
        return {
            "threshold": int(ask("ShotDetect Detection Threshold"))
        }
    except ValueError:
        return {}

# Packet Tagger
def packet_tagger_generator (plugin, ask):
    config = {}

    audio_source_id = ask("Audio Source ID")
    if audio_source_id:
        config['audioSourceId'] = audio_source_id

    video_source_id = ask("Video Source ID")
    if video_source_id:
        config['videoSourceId'] = video_source_id

    return config

# Packet Filter
packet_filters = {
    "Allow Video Only": {
        "allowIf": {
            "type": "video"
        }
    },
    "Allow Audio Only": {
        "allowIf": {
            "type": "audio"
        }
    }
}
def packet_filter_generator (plugin, ask):
    config = {}
    filter_type = ask("Packet Filter Type", packet_filters.keys())
    if filter_type and filter_type in packet_filters:
        config.update(packet_filters[filter_type])
    return config

known_plugins = dict([(plugin.id, plugin) for plugin in [
    KnownPlugin(["root"], "forestgreen"),
    KnownPlugin(["cattura.capturing.java_sound.JavaSoundCapturePlugin"], "darkseagreen", audio_capture_mouseover, audio_capture_generator),
    KnownPlugin(["cattura.capturing.alsa.ALSACapturePlugin"], "darkseagreen", audio_capture_mouseover, audio_capture_generator),
    KnownPlugin(["cattura.capturing.ndi.NDIDaemonCapturePlugin"], "darkseagreen", device_capture_mouseover),
    KnownPlugin(["BlackMagicVideoCapturePlugin"], "firebrick", video_capture_mouseover, video_capture_generator),
    KnownPlugin(["cattura.capturing.video4linux.V4LCapturePlugin"], "firebrick", video_capture_mouseover, video_capture_generator),
    KnownPlugin(["cattura.processing.resampler.XugglerVideoResamplerPlugin"], "yellow", resampler_mouseover),
    KnownPlugin(["cattura.processing.compositor.FrameCompositorPlugin"], "fuchsia", stitcher_mouseover),
    KnownPlugin(["cattura.encoding.writer.XuggleWriterPlugin"], "cadetblue", writer_mouseover),
    KnownPlugin(["cattura.encoding.transmux.FFmpegTransmuxerPlugin"], "cadetblue", transmuxer_mouseover),
    KnownPlugin(["cattura.encoding.qtfaststart.QTFaststartPlugin"], "lightblue"),
    KnownPlugin(["cattura.processing.shotdetect.ShotDetectPlugin"], "coral", shotdetect_mouseover, shotdetect_generator),
    KnownPlugin(["cattura.processing.tesseract.TesseractOCRPlugin"], "bisque"),
    KnownPlugin(["cattura.metadata.element_supplier.ElementSupplierPlugin", "MetadataPlugin"], "plum", metadata_mouseover),
    KnownPlugin(["cattura.capturing.file_reader.FileReaderPlugin"], "cyan", file_reader_mouseover, file_reader_generator),
    KnownPlugin(["LocalStoragePlugin"], "darkolivegreen"),
    KnownPlugin(["NoSignalFrameFilterPlugin"], "silver", no_signal_mouseover),
    KnownPlugin(["CWSStreamerPlugin"], "mediumslateblue", cws_mouseover),
    KnownPlugin(["cattura.encoding.save_frame.SaveFramePlugin"], "indigo", save_frame_mouseover),
    KnownPlugin(["cattura.capturing.text_generator.TextGeneratorPlugin"], "linen", text_generator_mouseover),
    KnownPlugin(["cattura.processing.fps_throttle.FPSThrottlePlugin"],  "hotpink", fps_mouseover),
    KnownPlugin(["cattura.metadata.external_asset.AssetPlugin"], "lightcyan", metadata_mouseover, asset_generator),
    KnownPlugin(["cattura.analysis.packet_analysis.PacketAnalysisPlugin", "com.cattura.plugins.analysis.chart.PacketChartPlugin"], "burlywood"),
    KnownPlugin(["cattura.analysis.audio_dumper.AudioDumperPlugin"], "tan"),
    KnownPlugin(["cattura.processing.packet_tagger.PacketTaggerPlugin"], "darkcyan", None, packet_tagger_generator),
    KnownPlugin(["cattura.processing.packet_filter.PacketFilterPlugin"], "aquamarine", None, packet_filter_generator)
]])

def find_plugin (plugin_id):
    plugin = known_plugins.get(plugin_id, None)
    if plugin:
        return plugin

    for key, plugin in known_plugins.iteritems():
        if key == plugin_id or plugin.simple_id == plugin_id or plugin_id in plugin.also_known_as:
            return plugin

    if "." in plugin_id:
        simple_plugin_id = plugin_id.split(".")[-1]
        return find_plugin(simple_plugin_id)
