# vim: set fileencoding=utf-8 :(
from __future__ import division
from __future__ import print_function

import sys
import os
import os.path
import json
import collections
import traceback
import tempfile
import subprocess
import threading
import getpass

from zipfile import ZipFile
from datetime import datetime
from math import atan, sin, cos, pi, sqrt, acos, atan2, asin

from PySide import QtCore, QtGui
from PySide.QtCore import QPointF, QSettings, QByteArray
from PySide.QtCore import Qt, QTimer

from PySide.QtGui import QGraphicsTextItem, QFont, QGraphicsLineItem, QIcon
from PySide.QtGui import QMenu, QMainWindow, QGraphicsSimpleTextItem
from PySide.QtGui import QPainterPath, QGraphicsPathItem, QFileDialog
from PySide.QtGui import QMessageBox, QDockWidget, QTextEdit, QPlainTextEdit
from PySide.QtGui import QGridLayout, QWidget, QLabel, QLineEdit, QPushButton
from PySide.QtGui import QPalette, QColor, QSplitter, QGraphicsScene
from PySide.QtGui import QGraphicsView, QGraphicsItem, QHBoxLayout
from PySide.QtGui import QInputDialog

from PySide.QtSvg import QGraphicsSvgItem, QSvgRenderer

import gv
from pygraph.classes.graph import graph
from pygraph.readwrite.dot import write

from lxml import etree

import signal
signal.signal(signal.SIGINT, signal.SIG_DFL)

from . import graph_runner, log_viewer, config_tester, output_checker, terminal_exec
from .known_plugins import known_plugins, find_plugin
from .invoker import invoke_in_main_thread

SVG_PTH = os.path.join(os.path.dirname(__file__), 'svg', 'items.svg')

svgDocType = """<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN" 
  "http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd">"""

CO_NAME = "CatturaVideo"
APP_NAME = "CATGUT"

VERSION = "1.0.6"

GRAPH_JSON = "graph.json"

CONNECTED_TO = "connectedTo"
circleCollisionMargin = 12 

nsmap = {
    'sodipodi': 'http://sodipodi.sourceforge.net/DTD/sodipodi-0.dtd',
    'cc': 'http://web.resource.org/cc/',
    'svg': 'http://www.w3.org/2000/svg',
    'dc': 'http://purl.org/dc/elements/1.1/',
    'xlink': 'http://www.w3.org/1999/xlink',
    'rdf': 'http://www.w3.org/1999/02/22-rdf-syntax-ns#',
    'inkscape': 'http://www.inkscape.org/namespaces/inkscape'
    }

FILE_MANAGER = os.environ.get("CATGUT_FILE_MANAGER", "xdg-open")
TERMINAL_EMULATOR = os.environ.get("CATGUT_TERMINAL_EMULATOR", "x-terminal-emulator -e")
LOG_VIEWER_COMMAND = os.environ.get("CATGUT_LOG_VIEWER", "less")

terminal_executor = terminal_exec.TerminalExecutor(TERMINAL_EMULATOR.split(" "))
view_log = log_viewer.LogViewer(terminal_executor, LOG_VIEWER_COMMAND)
capture_runner = graph_runner.Runner(terminal_executor)
node_config_tester = config_tester.Tester(capture_runner)
node_output_checker = output_checker.OutputChecker(capture_runner)

class JSONCoercionError(Exception):
    def __init__(self, key, text, msg):
        self.key = key
        self.text = text
        Exception.__init__(self, "Could not coerce %s key to JSON. value: %r error: %s" % (key, text, msg))


class CatgutNodeEditor(QWidget):
    node = None     # currently "focused" Node, available for editing.

    def __init__(self, mainW, *args, **kw):
        super(CatgutNodeEditor, self).__init__()

        self.mainW = mainW

        self.newLayout()
        
        # map doc field name ("config" etc) to 2-tuple of (editWidget, fetchFunction)
        # where fetchFunction is passed a QLineEdit or QPlainTextEdit and will return
        # the unicode text of the widget.
        self.editorWidgets = {}

        current = self.mainW.nodes.get(0, None) 
        if current is not None:
            self.setFocusedNode(current)

    def newLayout(self):
        # delete current layout
        #
        # un-sets the layout, and re-parents the layout's widget to an object
        # that goes away immediately.
        QWidget().setLayout(self.layout())

        self.grid = QGridLayout()
        self.grid.setColumnMinimumWidth(3, 220)

        self.setLayout(self.grid)

    def commitEdits(self):
        for key, (editWidget, fetchFunction) in self.editorWidgets.items():
            text = fetchFunction()
            try:
                newData = json.loads( text )
            except Exception, e:
                print("Exception trying to save %r's JSON value of %r:" % (key, text))
                traceback.print_exc()
                raise JSONCoercionError(key, text, str(e))

            if key == CONNECTED_TO:
                fromNodeID = self.node.json["nodeID"]

                # set "difference" to determine which edges to delete.
                toDelete = set(self.node.json[CONNECTED_TO]) - set(newData)

                # set "difference" to determine which edges to add.
                toAdd = set(newData) - set(self.node.json[CONNECTED_TO])

                for toNodeID in toDelete:
                    self.mainW.deleteEdge( fromNodeID, toNodeID )

                for toNodeID in toAdd:
                    self.mainW.addEdge( fromNodeID, toNodeID )

            else:
                if key == "config":
                    # "config" is special. we parsed it as newData just to be sure
                    # that it is actually valid json. but it is
                    # supposed to appear in the document as a string (nested JSON).
                    self.node.json[key] = text
                    self.configChanged(newData) # update our color, etc
                else:
                    self.node.json[key] = newData

    def removeConnectedTo(self, toNodeID):
        """
        Remove a NodeID from the connectedTo list, if possible.
        """
        key = CONNECTED_TO
        editWidget, fetchFunction = self.editorWidgets[key]
        text = fetchFunction()
        try:
            newData = json.loads( text )
            if toNodeID in newData:
                newData.remove(toNodeID)

            newText = json.dumps(newData)
            editWidget.setText(newText)
        except Exception, e:
            print("Exception trying to parse %r's JSON value of %r:" % (key, text))
            traceback.print_exc()

    def addConnectedTo(self, toNodeID):
        """
        Add a NodeID to the connectedTo list, if possible.
        """
        key = CONNECTED_TO
        editWidget, fetchFunction = self.editorWidgets[key]
        text = fetchFunction()
        try:
            newData = json.loads( text )
            if toNodeID not in newData:
                newData.append(toNodeID)

            newText = json.dumps(newData)
            editWidget.setText(newText)
        except Exception, e:
            print("Exception trying to parse %r's JSON value of %r:" % (key, text))
            traceback.print_exc()


    def configChanged(self, config):
        """
        JSON config changed.
        """
        if 'pluginID' in config:
            pluginID = config["pluginID"]

            plugin = find_plugin(pluginID)
            loadSvgCached(self.node, SVG_PTH, "default-node", fill=plugin.color if plugin else "none")

    def tryCommitEdits(self):
        try:
            self.commitEdits()
        except JSONCoercionError, e:
            self.errLabel.setText("Bad JSON value for \"%s\"." % (e.key,))
            # don't allow focus to change until edit error is cleared.
            return False
        else:
            # If no node focused by editor yet, then may not have this widget...
            errLabel = getattr(self, 'errLabel', None)
            if errLabel is not None:
                errLabel.setText(u"") # clear any previous error

        return True


    def setFocusedNode(self, n, undoExisting=False):
        if n is self.node and not undoExisting:
            return

        # first record any changes 
        if self.node and not undoExisting:
            if not self.tryCommitEdits():
                return

        if self.node:
            self.newLayout()

        self.node = n

        i = 0
        lbl = QLabel( u"{ " )
        self.grid.addWidget(lbl, i, 0)

        i += 1

        for key, value in n.json.items():
            lbl = QLabel( u"\"%s\"" % (key,), parent=self)
            self.grid.addWidget(lbl, i, 1)

            lbl = QLabel( u" : ", parent=self)
            self.grid.addWidget(lbl, i, 2)


            if key == "config":
                # format the config text nicely and put it in a multi-line edit box.
                i += 1
                edit = QPlainTextEdit(parent=self)
                try:
                    text = json.dumps( json.loads(value), indent=4, sort_keys=True, separators=(',', ': '))
                except ValueError:
                    text = value
                edit.setPlainText(text)
                self.grid.addWidget(edit, i, 1, 1, 3)

                self.editorWidgets[key] = (edit, edit.toPlainText)
            else:
                #if isinstance(value, (str, unicode)):
                #    edit = QLineEdit(parent=self)
                #    self.grid.addWidget(edit, i, 2)
                edit = QLineEdit(parent=self)
                edit.setText( json.dumps(value) )
                self.grid.addWidget(edit, i, 3)

                self.editorWidgets[key] = (edit, edit.text)

            i += 1

        i += 1
        lbl = QLabel( u"} " )
        self.grid.addWidget(lbl, i, 0)

        i += 1

        # button row 1
        hbox = QWidget()
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)

        btn = QPushButton(u"Commit edits", parent=self)
        btn.clicked.connect(self.tryCommitEdits)
        layout.addWidget(btn)

        btn = QPushButton(u"Undo changes", parent=self)
        btn.clicked.connect(self.undoChanges)
        layout.addWidget(btn)

        hbox.setLayout(layout)
        self.grid.addWidget(hbox, i, 0, 1, 4)

        i += 1

        # button row 2
        hbox = QWidget()
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)

        btn = QPushButton(u"View Processor Log", parent=self)
        btn.clicked.connect(self.viewProcessorLog)
        btn.setEnabled(n.log_file is not None)
        layout.addWidget(btn)

        hbox.setLayout(layout)
        self.grid.addWidget(hbox, i, 0, 1, 4)

        i += 1

        # button row 2
        hbox = QWidget()
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)

        btn = QPushButton(u"Test Plugin Configuration", parent=self)
        btn.clicked.connect(self.testPluginConfiguration)
        btn.setEnabled(capture_runner.is_usable())
        layout.addWidget(btn)

        config = {"type": ""}
        try:
            config = json.loads(n.json['config'])
        except ValueError: pass

        btn = QPushButton(u"Test Authentication", parent=self)
        btn.clicked.connect(self.testAuthentication)
        btn.setEnabled(capture_runner.is_usable() and config.get("type", "") == "INTEGRATION")
        layout.addWidget(btn)

        hbox.setLayout(layout)
        self.grid.addWidget(hbox, i, 0, 1, 4)

        i += 1
        
        # button row 3
        hbox = QWidget()
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)

        btn = QPushButton(u"Get Expected Output of Type", parent=self)
        btn.clicked.connect(self.getNodeOutputOfType)
        btn.setEnabled(capture_runner.is_usable())
        layout.addWidget(btn)

        btn = QPushButton(u"View Output JSON", parent=self)
        btn.clicked.connect(self.viewOutputJson)
        btn.setEnabled(n.processor_directory is not None)
        layout.addWidget(btn)

        hbox.setLayout(layout)
        self.grid.addWidget(hbox, i, 0, 1, 4)

        i += 1

        # button row 4
        hbox = QWidget()
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)

        btn = QPushButton(u"Open Processor Directory", parent=self)
        btn.clicked.connect(self.openProcessorDirectory)
        btn.setEnabled(n.processor_directory is not None)
        layout.addWidget(btn)

        hbox.setLayout(layout)
        self.grid.addWidget(hbox, i, 0, 1, 4)

        i += 1

        self.errLabel = QLabel(u"")
        p = QPalette(self.mainW.app.palette())
        p.setColor(QPalette.WindowText, QColor("red"))
        p.setColor(QPalette.Text, QColor("red"))
        self.errLabel.setPalette(p)
        self.grid.addWidget(self.errLabel, i, 0, 1, 4)

        i += 1
        # stretch the final (empty) row to fill up excess space
        #self.grid.setRowStretch(i, 1)


    def undoChanges(self):
        self.setFocusedNode(self.node, undoExisting=True)

    def viewProcessorLog(self):
        view_log(self.node.log_file)

    def openProcessorDirectory(self):
        if self.node.processor_directory is None:
            return
        subprocess.call([FILE_MANAGER, self.node.processor_directory])

    def viewOutputJson(self):
        try:
            result = node_output_checker.get_output_json(self.node)
            QMessageBox.information(self, "Get Output Result", json.dumps(result, indent=4))
        except ValueError as error:
            QMessageBox.information(self, "Parse error", "Node output is invalid JSON: " + error.message)    

    def testPluginConfiguration(self):
        if not do_xuggler_directory_test(self):
            return

        config = None
        message = "Node configuration is invalid JSON"
        try:
            config = json.loads(self.editorWidgets["config"][1]())
            result = node_config_tester.test_config_structure(self.node, config)
            message = "Node configuration invalid"
            if result:
                message = "Node configuration valid"
        except ValueError as error:
            message = "Node configuration is invalid JSON: " + error.message

        QMessageBox.information(self, "Test Configuration Result", message)
    
    def testAuthentication(self):
        if not do_xuggler_directory_test(self):
            return

        config = None
        message = "Node configuration is invalid JSON"
        try:
            config = json.loads(self.editorWidgets["config"][1]())
            result = node_config_tester.test_authentication(self.node, config)
            print(result)
            message = result["type"]
            if result.get("errorMessage", ""):
                message = message + ": " + result["errorMessage"]

            if result.get("exception", ""):
                message = message + "\n\n" + dumpCaptureCastException(result["exception"])
        except ValueError as error:
            message = "Node configuration is invalid JSON: " + error.message
        
        QMessageBox.information(self, "Authentication Result", message)

    def getNodeOutputOfType(self):
        if not do_xuggler_directory_test(self):
            return

        try:
            config = json.loads(self.editorWidgets["config"][1]())
            (output_type, ok_pressed) = QInputDialog.getItem(self, "Get Output Of Type", "Type", [
                "Asset", "Video Frame", "Metadata", "Audio Format",
                "Encoded Video Packet", "Encoded Audio Packet",
                "Table of Contents", "Table of Contents Entry",
                "Scene Change", "Scene Change Entry", "Transcript"
            ])
            if not ok_pressed:
                return

            result = node_output_checker.get_output_of_type(self.mainW.json, self.node, output_type, config)
            QMessageBox.information(self, "Get Output Result", json.dumps(result, indent=4))
        except ValueError as error:
            QMessageBox.information(self, "Parse error", "Node configuration is invalid JSON: " + error.message)

def dumpCaptureCastException (exception):
    if not exception:
        return ""

    result = [exception["message"]]

    for stackTraceElement in exception["stackTrace"]:
        result.append("  at {className}.{methodName}({fileName}:{lineNumber})".format(**stackTraceElement))

    if exception.get("cause", None):
        result.append("Caused by: " + dumpCaptureCastException(exception["cause"]))

    return "\n".join(result)

class CatgutColorkeyWidget(QGraphicsView):
    """
    Provides a "map key" for the main graphics scene showing node colors
    for certain known plugin id's
    """
    def __init__(self, *args, **kw):
        self.scene = QGraphicsScene()
        itemY = 0

        cScale = 0.45
        for pluginID, plugin in known_plugins.items():
            click_handler = lambda plugin: scene.newNode(None, self.makePluginConfig(plugin))
            circle = PluginLegendCircle(plugin, click_handler)

            circle.scale(cScale, cScale)

            rect = circle.boundingRect()
            text = PluginIdLabel(plugin, click_handler)

            circle.setX(0)
            circle.setY(itemY)

            text.setX(rect.width() * cScale)
            text.setY(itemY + rect.height() * cScale / 2 - 11)

            self.scene.addItem(circle)
            self.scene.addItem(text)

            itemY += rect.height() * cScale

        self.height = itemY + 10

        super(CatgutColorkeyWidget, self).__init__(self.scene, *args, **kw)

    def makePluginConfig(self, plugin):
        try:
            return plugin.make_config(lambda question, choices=None: self.ask(question, choices))
        except Exception, e:
            QMessageBox.critical(self, "Error creating node", str(e))
            raise e

    def ask(self, question, choices=None):
        retval = None
        if choices:
            retval = QInputDialog.getItem(self, "Plugin Configuration", question, choices)
        else:
            retval = QInputDialog.getText(self, "Plugin Configuration", question)
        
        if retval[1]:
            return retval[0]

    #def mouseDoubleClickEvent(self, event):
    #    print(event)


class NodeHoverTextBoxItem(QGraphicsTextItem):
    def __init__(self, node, scene, text, width, *args, **kw):
        super(NodeHoverTextBoxItem, self).__init__(*args, **kw)

        self.node = node

        self.setHtml(self.prettyText(text))

        width = self.boundingRect().width() + 5
        height = self.boundingRect().height() + 4

        box = QGraphicsSvgItem()
        box.setParentItem(self)
        box.setFlag(QGraphicsItem.ItemStacksBehindParent, True)
        box.setX(-5)

        loadSvgCached(box, SVG_PTH, "textbox", width=width, height=height)

        scene.addItem(self)

    def prettyText(self, text):
        lines = text.split('\n')
        for i, line in enumerate(lines):
            parts = line.split(':', 1)
            if len(parts) == 2:
                lines[i] = '<b>' + parts[0] + ':</b>' + parts[1]

        return '<br>'.join(lines)

    def hoverEnterEvent(self, event):
        if self.node.hoverTimer is not None:
            # cancel timeout since we are hovering the mouseover text itself,
            # which should not result in the mouseover text being timed out and removed from the scene.
            self.node.hoverEnterEvent(event)

    def hoverLeaveEvent(self, event):
        if self.node.hoverTimer is None:
            # if there is no timeout for the hover text, then we should start one
            # since we've moused off of the hover text.
            self.node.hoverLeaveEvent(event)

def getMouseOverText(nodeID, pluginID, config):
    text = '<b>'+pluginID+'</b>\n\n'

    try:
        plugin = find_plugin(pluginID)
        if plugin:
            text += plugin.mouseover_text(config)

    except Exception, e:
        text += '<error getting mouseover text, see stderr>'
        print('Exception getting mouseover text for node %s' % (nodeID,), file=sys.stderr)
        traceback.print_exc()
    return text

def do_xuggler_directory_test (window):
    result = graph_runner.test_xuggler_directory()
    if not result:
        QMessageBox.warning(
            window,
            "Xuggler Directory Not Writable",
            "Xuggler directory ({}) not writable by current user ({}). CaptureCast will not run.".format(
                graph_runner.XUGGLER_DIRECTORY, getpass.getuser()
            )
        )
    return result

class CatgutMainWindow(QMainWindow):
    filename = None
    graphFilesFilter = "Graph files  *.json (*.json);;Log dumps *.zip (*.zip);;All files  *.* (*)"
    json = None # full JSON data structure for current document.

    nodes = None

    def __init__(self, app, *args, **kw):
        QMainWindow.__init__(self, *args, **kw)

        self.app = app
        self.nodes = {}

        menuBar = self.menuBar()
        fileMenu = menuBar.addMenu("&File")

        a = fileMenu.addAction("&Open Graph")
        a.triggered.connect(self.openFile)

        a = fileMenu.addAction("&New Graph")
        a.triggered.connect(self.newGraph)
        fileMenu.addSeparator()

        a = fileMenu.addAction("&Save")
        a.triggered.connect(self.saveFile)

        a = fileMenu.addAction("Sa&ve As")
        a.triggered.connect(self.saveFileAs)

        fileMenu.addSeparator()
        a = fileMenu.addAction("&Quit")
        a.triggered.connect(self.quit)

        testMenu = menuBar.addMenu("&Test")
        testMenu.setEnabled(capture_runner.is_usable())

        a = testMenu.addAction("&Test Graph")
        a.triggered.connect(self.testGraph)

        helpMenu = menuBar.addMenu("&Help")
        a = helpMenu.addAction("&About")
        a.triggered.connect(self.about)

        helpMenu.addSeparator()
        a = helpMenu.addAction("About &Qt")
        a.triggered.connect(self.aboutQt)

        status = self.statusBar()


    def closeEvent(self, event):
        settings = QSettings(CO_NAME, APP_NAME)
        settings.setValue("geometry", self.saveGeometry())
        settings.setValue("windowState", self.saveState())
        QMainWindow.closeEvent(self, event)


    def readSettings(self):
        settings = QSettings(CO_NAME, APP_NAME)
        self.restoreGeometry(settings.value("geometry"))
        self.restoreState(settings.value("windowState"))


    def deleteGraph(self):
        self.json = {"nodes":[]}
        for c in list(self.nodes.values()):
            # should delete edges also and remove GUI elements from the graphics scene
            self.deleteNode(c)


    def newGraph(self):
        self.deleteGraph()

        node = {"config": "{\"pluginID\":\"root\",\"type\":\"READ\",\"baseDirectory\":\"/var/cattura/storage/templates/Some-Template\"}",
                "connectedTo": [],
                "nodeID": 0,
                "workingDir": "/var/cattura/storage/templates/Some-Template"
            }

        self.json = {"nodes" : [node]}

        plugin = find_plugin("root")
        n = Node(self, 0, plugin.color if plugin else "none", node, None, None)
        n.setX(0)
        n.setY(0)

        self.editor.setFocusedNode(n)
    

    def about(self):
        about = u"CATGUT - the CATtura Graph Unfucking Tool.\n\n© 2013 - 2020 Cattura Video.\n\nMay causing inflamation and/or genital swelling. May not be combined with alcohol. Reports of superhuman strength are unconfirmed. Consult your Doctor before discontinuing use or adjusting dosage. Full money-back guarantee if less than completely alive. Offer not valid in Nevada. "
        QMessageBox.about(self, "About CATGUT", about)


    def aboutQt(self):
        QMessageBox.aboutQt(self)


    def deleteNode(self, n):
        del self.nodes[n.nodeID]

        # delete node from our JSON document.
        for i, node in enumerate(self.json["nodes"]):
            if node["nodeID"] == n.nodeID:
                print("Deleted Node %d from JSON document." % (n.nodeID,))
                del self.json["nodes"][i]

        for edge in edges.copy():
            if n is edge.headHandle.fromNode:
                edge.headHandle.deleteEdge()

            if n is edge.tailHandle.fromNode:
                edge.tailHandle.deleteEdge()

        scene.removeItem(n)
    

    def deleteEdge(self, fromNodeID, toNodeID, updateGUI=True):
        if updateGUI:
            for edge in edges.copy():
                if edge.headHandle.fromNode.nodeID == fromNodeID and edge.tailHandle.fromNode.nodeID == toNodeID:
                    edges.remove(edge)

                    scene.removeItem(edge)
                    scene.removeItem(edge.headHandle)
                    scene.removeItem(edge.tailHandle)

        #print("deleteEdge( %d, %d )" % (fromNodeID, toNodeID))
        for node in self.json["nodes"]:
            if node["nodeID"] == fromNodeID:
                print("Removed Edge from Node %r to Node %r from JSON document." % (fromNodeID, toNodeID))
                node["connectedTo"].remove(toNodeID)

                # also update the Node Editor content if we're editing the "from" node currently. otherwise
                # when the Node Editor commits its changes to the document, it will recreate edge connections
                # according to it's own "connectedTo" list.
                if self.editor.node is not None and self.editor.node.nodeID == fromNodeID:
                    self.editor.removeConnectedTo(toNodeID)


    def addEdge(self, fromNodeID, toNodeID, updateGUI=True):
        if updateGUI:
            e = Edge(self, self.nodes[fromNodeID], self.nodes[toNodeID])
            edges.add(e)
            scene.addItem(e)

        #print("addEdge( %d, %d )" % (fromNodeID, toNodeID))
        for node in self.json["nodes"]:
            if node["nodeID"] == fromNodeID:
                print("Added Edge from Node %r to Node %r to JSON document." % (fromNodeID, toNodeID,))
                if toNodeID not in node["connectedTo"]:
                    node["connectedTo"].append(toNodeID)

                # also update the Node Editor content if we're editing the "from" node currently. otherwise
                # when the Node Editor commits its changes to the document, it will delete edge connections
                # according to it's own "connectedTo" list.
                if self.editor.node is not None and self.editor.node.nodeID == fromNodeID:
                    self.editor.addConnectedTo(toNodeID)


    def quit(self):
        self.app.closeAllWindows()
        # Qt should quit after last window is closed.


    def openFile(self):
        filename, filter = QFileDialog.getOpenFileName(self, "Open Graph", ".", self.graphFilesFilter)

        if filename:
            self.filename = filename
            self.doOpenFile()


    def saveFile(self):
        if not self.editor.tryCommitEdits():
            return

        if self.filename:
            self.doSaveFile()
        else:
            self.saveFileAs()


    def saveFileAs(self):
        if not self.editor.tryCommitEdits():
            return

        filename, filter = QFileDialog.getSaveFileName(self, "Save Graph As", ".", self.graphFilesFilter)
        #print("stuff", filename, filter)

        if filename:
            # Note: Dialog has already confirmed overwrite of any existing file.
            self.filename = filename
            self.doSaveFile()


    def doSaveFile(self):
        print('doSaveFile', self.filename)

        try:
            with open(self.filename, 'wb') as f:
                json.dump(self.json, f, indent=4, sort_keys=True, separators=(',', ': '))
        except Exception, e:
            print("Error saving file:", file=sys.stderr)
            print(e, file=sys.stderr)
            QMessageBox.critical(self, "Error saving file", str(e))

    def doOpenFile(self, filename=None):
        if filename is not None:
            self.filename = filename
        
        # If we're given a zip file, unpack it to a temp directory
        # and select the graph.json inside it.
        if self.filename.endswith(".zip"):
            zip_source = ZipFile(self.filename)
            zip_target = tempfile.mkdtemp()
            zip_source.extractall(zip_target)
            self.filename = os.path.join(zip_target, GRAPH_JSON)

        # If we're given a directory, select the graph.json inside it.
        if os.path.isdir(self.filename):
            self.filename = os.path.join(self.filename, GRAPH_JSON)

        gr = graph()

        # load JSON graph file
        with open(self.filename, 'rb') as f:
            j = json.load(f)

        # VERIFY structure of document before loading GUI elements as we don't want to partially
        # create GUI elements for this document if it fails to parse half way through.
        try:
            nodes = j["nodes"]
        except KeyError, e:
            sys.stderr.write("Malformed JSON graph file. Missing \"nodes\" key. Aborting!\n")
            return
        for node in nodes:
            try:
                nodeID = node["nodeID"]
                connectedTo = node["connectedTo"]
                #pluginID = config["pluginID"]
            except Exception, e:
                print("Malformed JSON graph file. Missing \"nodeID\" and/or \"connectedTo\" keys and/or \"config\" didn't parse. Aborting!",
                      file=sys.stderr)
                print(e, file=sys.stderr)
                return

        # clear any existing document / gui elements.
        self.json = {"nodes":[]}
        for c in list(self.nodes.values()):
            # should delete edges also and remove GUI elements from the graphics scene
            self.deleteNode(c)


        # set our new document
        self.json = j

        find_log = log_viewer.LogFinder(self.filename)
        for node in nodes:
            nodeID = node["nodeID"]
            connectedTo = node["connectedTo"]

            try:
                config = json.loads(node["config"])
                pluginID = config["pluginID"]
            except:
                pluginID = "<no config>"

            plugin = find_plugin(pluginID)
            color = plugin.color if plugin else "none"

            Node(self, nodeID, color, node, find_log.find_processor_directory(node), find_log(node))

            gr.add_nodes([str(nodeID)])


        for node in self.nodes.values():
            for otherID in node.json["connectedTo"]:
                otherNode = self.nodes.get(otherID, None)
                if otherNode is None:
                    sys.stderr.write("Malformed JSON graph file. Node %d connected to %d, but %d does not exist. Aborting!\n", (node.nodeID, otherID, otherID))
                    sys.exit(1)

                e = Edge(self, node, otherNode)
                edges.add(e)
                scene.addItem(e)

                gr.add_edge( (str(node.nodeID), str(otherNode.nodeID)) )

        dot = write(gr)
        gvv = gv.readstring(dot)
        gv.layout(gvv,'dot')

        gv.render(gvv)

        for nodeID in self.nodes.keys():
            nodeIDStr = str(nodeID)

            n = gv.findnode(gvv, nodeIDStr)
            x,y = [round(float(X)) for X in gv.getv(n, "pos").split(',')]

            self.nodes[nodeID].setX(x*2)
            self.nodes[nodeID].setY(y*2)

        # not 100% sure why I need to do this again here, but its necessary for the handles
        # to be drawn exactly where they are supposed to be, after the nodes have
        # been layed out...
        for e in edges:
            e.headHandle.nodeMoved()
            e.tailHandle.nodeMoved()
    
        # Focus on root node
        for node in self.nodes.values():
            if node.nodeID == 0:
                self.editor.setFocusedNode(node)

    def testGraph(self):
        if not do_xuggler_directory_test(self):
            return
            
        test_graph_path = "catgut_test_{}".format(str(datetime.now()).replace(" ", "_"))

        def after_test():
            answer = QMessageBox.question(self, "Test Graph", "Test graph instance generated at {0}. Would you like to open the generated test graph in a new CATGUT instance?".format(test_graph_path), QMessageBox.Ok | QMessageBox.Cancel)
            if answer == QMessageBox.StandardButton.FirstButton:
                subprocess.Popen(
                    ["python", "-m", "catgut", test_graph_path],
                    shell=False,
                    stdin=None,
                    stdout=None,
                    stderr=None,
                    close_fds=True
                )            

        def do_test():
            capture_runner(self.json, test_graph_path)
            invoke_in_main_thread(after_test)
        
        threading.Thread(target=do_test).start()

class CatgutGraphicsScene(QGraphicsScene):
    # set to an instance of Edge, if we're supposed to be
    # drawing (moving the tail handle) of an Edge instance.
    #
    # This gets set via right-clicking on a Circle and
    # selecting "Edge..."

    drawingEdge = None

    def __init__(self, mainW, *args, **kw):
        super(CatgutGraphicsScene, self).__init__(*args, **kw)

        self.mainW = mainW


    def contextMenuEvent(self, event):
        #print 'contextMenuEvent', event
        menu = QMenu(view)
        a = menu.addAction("Node...")
        a.triggered.connect(lambda: self.newNode(event))
        menu.exec_(event.screenPos())

    def mouseMoveEvent(self, event):
        #print 'CatgutGraphicsScene.mouseMoveEvent', event
        if self.drawingEdge is not None:
            pos = event.scenePos()

            self.drawingEdge.tailHandle.setX(pos.x())
            self.drawingEdge.tailHandle.setY(pos.y())

            self.drawingEdge.headHandle.nodeMoved()
        else:
            # tell super class to handle it so that it propogates
            QtGui.QGraphicsScene.mousePressEvent(self, event)

    def mousePressEvent(self, event):
        #print 'CatgutGraphicsScene.mousePressEvent', event
        if self.drawingEdge is not None and event.button() == Qt.MouseButton.LeftButton:

            # Can we drop on top of a Circle? if so, hook up
            # this edge and add it to the `edges' data structure.
            if self.drawingEdge.tailHandle.tryDrop(event):
                # dropped on to a Circle successfully. Edge objects have been changed.
                # HOWEVER, since this was an edge-in-progress (ie. we were in the progress of
                # drawing this edge out, it has not yet been added to the global `edges' structure.
                # Add it now, since we've successfully connected its tail to the Circle it was
                # dropped on.
                edges.add(self.drawingEdge)
            else:
                # dropped in dead space - remove edge from scene and stop drawing it.
                scene.removeItem(self.drawingEdge)
                scene.removeItem(self.drawingEdge.headHandle)
                scene.removeItem(self.drawingEdge.tailHandle)

            self.drawingEdge = None # either way, we aren't drawing this edge anymore...
        else:
            # tell super class to handle it so that it propogates
            QtGui.QGraphicsScene.mousePressEvent(self, event)

    def newNode(self, event, config={}):
        nodeID = 0 if not self.mainW.nodes else max(self.mainW.nodes.keys())+1

        node = {"config"      : json.dumps(config),
                "workingDir"  : "/var/cattura/storage/templates/",
                "nodeID"      : nodeID,
                "connectedTo" : []}

        color = "none"
        plugin_id = config.get("pluginID", None)
        if plugin_id:
            plugin = find_plugin(plugin_id)
            if plugin:
                color = plugin.color
        c = Node(self.mainW, nodeID, color, node, None, None)

        # update JSON document
        self.mainW.json["nodes"].append(node)

        if event:
            pos = event.scenePos()
            c.setX(pos.x())
            c.setY(pos.y())


class NodeIDText(QGraphicsTextItem):
    def __init__(self, *args, **kw):
        QGraphicsTextItem.__init__(self, *args, **kw)

        font = QFont(QtGui.QApplication.font())
        font.setBold(True)
        font.setPointSize(18)
        self.setFont(font)


class NodeLabelText(QGraphicsTextItem):
    def __init__(self, *args, **kw):
        QGraphicsTextItem.__init__(self, *args, **kw)

        font = QFont(QtGui.QApplication.font())
        font.setBold(False)
        font.setPointSize(10)
        self.setFont(font)


class PluginLegendWidget(object):
    def __init__(self, plugin, click_handler):
        self.setToolTip(plugin.id)
        self.click_handler = click_handler
        self.plugin = plugin


class PluginIdLabel(NodeLabelText, PluginLegendWidget):
    def __init__(self, plugin, click_handler):
        NodeLabelText.__init__(self, plugin.simple_id)
        PluginLegendWidget.__init__(self, plugin, click_handler)

    def mouseDoubleClickEvent(self, event):
        self.click_handler(self.plugin)


class PluginLegendCircle(QGraphicsSvgItem, PluginLegendWidget):
    def __init__(self, plugin, click_handler):
        QGraphicsSvgItem.__init__(self)
        PluginLegendWidget.__init__(self, plugin, click_handler)
        loadSvgCached(self, SVG_PTH, "default-node", fill=plugin.color)

    def mouseDoubleClickEvent(self, event):
        self.click_handler(self.plugin)

class Edge(QGraphicsPathItem):
    mainW = None

    def __repr__(self):
        return "<Edge %r --> %r>" % (self.headHandle.fromNode, self.tailHandle.fromNode)

    def __init__(self, mainW, fromNode, toNode, *args, **kw):
        QGraphicsPathItem.__init__(self, *args, **kw)

        self.mainW = mainW

        self.headHandle = Handle(self.mainW, self, fromNode, toNode, head=True)
        self.tailHandle = Handle(self.mainW, self, toNode, fromNode, head=False)

        self.headHandle.otherHandle = self.tailHandle
        self.tailHandle.otherHandle = self.headHandle

        scene.addItem(self.headHandle)
        scene.addItem(self.tailHandle)

        self.headHandle.xChanged.connect(self.handlesMoved)
        self.headHandle.yChanged.connect(self.handlesMoved)

        self.tailHandle.xChanged.connect(self.handlesMoved)
        self.tailHandle.yChanged.connect(self.handlesMoved)

        self.headHandle.nodeMoved() # update position
        self.tailHandle.nodeMoved() # update position

        # now update positions again, taking in to account changes in the "other" handle's position.
        self.headHandle.nodeMoved()
        self.tailHandle.nodeMoved()

        self.handlesMoved()


    def handlesMoved(self):
        """
        Draw spline between handles avoiding colliding nodes.
        """

        #print 'Edge.handlesMoved() %r'  % (self,)

        self.x1 = self.headHandle.x() + self.headHandle.boundingRect().width()/2
        self.y1 = self.headHandle.y() + self.headHandle.boundingRect().height()/2
        self.x2 = self.tailHandle.x()
        self.y2 = self.tailHandle.y()


        qppath = QPainterPath()
        #route = self.getBestEdgeRoute(self.headHandle.circle1.centerX, self.headHandle.circle1.centerY,
        #                              self.tailHandle.circle1.centerX, self.tailHandle.circle1.centerY)
        #if route and len(route.points) > 4: # must have at least 3 points to draw a spline...
        #    print 'here', route.points
        #    drawSpline(qppath, route.points, 0.5)
        #
        #else: # otherwise just draw a straight line...
            #print 'here2'
            #qppath.moveTo(self.headHandle.circle1.centerX, self.headHandle.circle1.centerY)
            #qppath.lineTo(self.tailHandle.circle1.centerX, self.tailHandle.circle1.centerY)

        qppath.moveTo(self.x1, self.y1)
        qppath.lineTo(self.x2, self.y2)

        self.setPath(qppath)


    def getBestEdgeRoute(self, fromX, fromY, toX, toY):
        #print 'getBestEdgeRoute'
        if fromX == toX and fromY == toY:
            print('skipping')
            return

        all_routes = []

        circs = [c for c in self.mainW.nodes.values() if c not in (self.headHandle.fromNode, self.tailHandle.fromNode)]

        route = Route()
        finishRoutesFrom(circs, all_routes, route, fromX, fromY, toX, toY)

        #for r in all_routes:
        #    print 'Route: %r length=%.2f' % (r.points, r.length)

        if all_routes:
            all_routes.sort(key=lambda r: r.length)
            return all_routes[0]


    def findCollidingCircles(self):
        if self.x1 == self.x2 and self.y1 == self.y2:
            return []
        circs = [c for c in self.mainW.nodes.values() if c not in (self.headHandle.fromNode, self.tailHandle.fromNode)]
        return findCollidingCircles(self.x1, self.y1, self.x2, self.y2, circs)


    def drawLine(self, x1, y1, x2, y2, text):
        i = QGraphicsSimpleTextItem(text)
        i.setX( x1 + (x2 - x1) / 2 )
        i.setY( y1 + (y2 - y1) / 2 )
        scene.addItem(i)

        line = QGraphicsLineItem()
        line.setLine(x1, y1, x2, y2)
        scene.addItem(line)

    def drawAngel(self, x1, y1, text, angel):
        i = QGraphicsSimpleTextItem(u'%s = °%.2f' % (text, angel * (180/pi)))
        i.setX( x1 )
        i.setY( y1 )
        scene.addItem(i)


def finishRoutesFrom(nodes, all_routes, base_route, x, y, endX, endY, ignoreCircle=None):
    route = Route()
    route.points = base_route.points[:]
    if base_route.points:
        route.length = base_route.length + distanceXY(base_route.points[-2], base_route.points[-1], x, y)

    #drawX(x,y)

    route.points += [x,y]

    # avoid infinite recursion where we keep colliding with the same circle
    # TODO optimize this
    #
    # TODO this is busted, i think 
    if ignoreCircle:
        tempNodes = nodes[:]
        if ignoreCircle in nodes:
            tempNodes.remove(ignoreCircle)
    else:
        tempNodes = nodes

    c = findFirstCollidingCircle(x, y, endX, endY, tempNodes)
    if c:
        #if distanceXY(x, y, c.centerX, c.centerY) <= c.radius + circleCollisionMargin:
        #    #print 'wtf'
        #    return

        data = findTangentPointsToCircle(x, y, c.centerX, c.centerY, c.radius + circleCollisionMargin)
        if data is None:
            #print 'wtf 2'
            return
        tanX,tanY,tanX2,tanY2 = data
        

        finishRoutesFrom(circles, all_routes, route, tanX, tanY, endX, endY, ignoreCircle=c) # branch "left"
        finishRoutesFrom(circles, all_routes, route, tanX2, tanY2, endX, endY, ignoreCircle=c) # branch "right"
        
    else:
        route.points += [endX, endY]
        route.length += distanceXY(x, y, endX, endY)
        all_routes.append(route) # route complete - remember it.


class Route(object):
    def __init__(self):
        self.points = []
        self.length = 0


def findFirstCollidingCircle(x1, y1, x2, y2, circs):
    circs = findCollidingCircles(x1, y1, x2, y2, circs)
    if circs:
        return circs[0]
    return None


def findCollidingCircles(x1, y1, x2, y2, circs):
    hit = []
    for c in circs:
        cRect = c.boundingRect()
        cr = cRect.width() / 2 + circleCollisionMargin
        cx = c.x() + cr
        cy = c.y() + cr
        #if x1 == x2 and y1 == y2:
        #    continue
        x,y = getPointPerpendicularFrom(x1, y1, x2, y2, cx, cy)

        # perpendicular point is somewhere along the line segment? (could be beyond either end... in which case it's
        # not indicative of a collision.

        # make sure point is between our endpoints...
        if x < max(x1, x2) and x > min(x1, x2) and y < max(y1, y2) and y > min(y1, y2):

            L = sqrt( (cx - x) ** 2 + (cy - y) ** 2 )
            #print 'L', L
            #print 'cr', cr

            # only if radius of circle is greater than perpendicular distance are we touching the line
            if cr > L:
                hit.append(c)

    # TODO
    # sort by closeness to starting end of edge?
    hit.sort(key=lambda c: -distanceToCircleCenter(x1, y1, c))
    return hit


def distanceToCircleCenter(x1, y1, c):
    cRect = c.boundingRect()
    cr = cRect.width() / 2
    cx = c.x() + cr
    cy = c.y() + cr

    return sqrt( (cx-x1)**2 + (cy-y1)**2 )


def getPointPerpendicularFrom(x1, y1, x2, y2, cx, cy):
    #if x1 == x2 and y1 == y2:
    #    return (x1,y1)

    k = ((y2 - y1) * 
           (cx - x1) - 
           (x2 - x1) * 
           (cy - y1)) / ((y2 - y1) ** 2 +
                              (x2 - x1) ** 2)
    px = cx - k * (y2 - y1)
    py = cy + k * (x2 - x1)

    #line = QGraphicsLineItem()
    #line.setLine(px, py, cx, cy)
    #scene.addItem(line)

    return (px, py)



renderers = {}
def loadSvgCached(svgItem, svgPath, elementId, **kw):

    cacheKey = (svgPath, elementId, tuple([ (k,v) for k,v in kw.items()]))
    r = renderers.get(cacheKey, None)
    if r is None:
        r = renderers[cacheKey] = QSvgRenderer()

        with open(svgPath, 'rb') as f:
            doc = etree.fromstring(f.read())

        p = '//svg:*[@id="%s"]' % (elementId,) # TODO a better way to getElementById() ???
        e = doc.xpath(p, namespaces=nsmap)[0]
        for attr, value in kw.items():
            e.set(attr, str(value))

        content = etree.tostring(doc, xml_declaration=True,
                                 standalone=False,
                                 doctype=svgDocType,
                                 encoding="utf-8")
        r.load(QByteArray(content))

    svgItem.setSharedRenderer(r)
    svgItem.setElementId(elementId)



class Node(QGraphicsSvgItem):
    mainW   = None

    radius  = None
    centerX = None
    centerY = None

    # Cattura info
    json = None # Python datastructure representing the parsed "node" object from the graph file.

    processor_directory = None # Directory where the processor associated to this node lives.
    log_file = None # Path to log file on disk.

    hoverText = None # set to instance of NodeHoverTextBoxItem when hovering
                     # over item - displays mouseover text info.
    hoverTimer = None # mouseover timer event to remove mouseover text

    def __repr__(self):
        return "<Circle nodeID=%d>" % (self.nodeID,)

    def __init__(self, mainW, nodeID, color, jsonData, processor_directory, log_file, *args, **kw):
        super(Node, self).__init__(*args, **kw)

        loadSvgCached(self, SVG_PTH, "default-node", fill=color)

        self.setAcceptHoverEvents(True)

        self.mainW = mainW
        self.nodeID = nodeID
        self.json = jsonData
        self.processor_directory = processor_directory
        self.log_file = log_file
        self.nodeIDStr = str(nodeID)

        self.mainW.nodes[nodeID] = self
        scene.addItem(self)

        self.text = NodeIDText(self.nodeIDStr, parent=self)
        if len(self.nodeIDStr) <= 1:
            self.text.setX(24)
        else:
            self.text.setX(15)
        self.text.setY(20)

        ourWidth = self.boundingRect().width()

        #self.labelText = NodeLabelText(self.label, parent=self)
        #labelWidth = self.labelText.boundingRect().width()

        #self.labelText.setX( ourWidth / 2 - labelWidth / 2 )
        #self.labelText.setY(40)

        self.xChanged.connect(self.updateCenterCoords)
        self.yChanged.connect(self.updateCenterCoords)

        self.updateCenterCoords()

    def connectedTo(self, *otherItems):
        self.connectedTo = otherItems

        for line in self.linesTo:
            scene.removeItem(line)

        for item in otherItems:
            line = scene.addLine(self.x(), self.y(), item.x(), item.y())
            self.linesTo.append(line)

    def getMouseOverText(self):
        try:
            config = json.loads(self.json["config"])
            pluginID = config["pluginID"].split('.')[-1]
            mousetext = getMouseOverText(self.nodeID, pluginID, config)
        except:
            mousetext = "<no \"config\" data, or misformatted>"
        return mousetext
 
    def hoverEnterEvent(self, event):
        if self.hoverTimer is not None:
            self.hoverTimer.stop()
            self.hoverTimer = None
            return

        text = self.getMouseOverText()
        width = 210
        self.hoverText = NodeHoverTextBoxItem(self, scene, text, width)

        self.hoverText.setX(self.x() + self.boundingRect().width() - 30)
        self.hoverText.setY(self.y() + self.boundingRect().height() - 22)

    def hoverLeaveEvent(self, event):
        self.hoverTimer = t = QTimer()
        t.setSingleShot(True)
        t.start(350)
        t.timeout.connect(self.removeHoverText)

    def removeHoverText(self):
        self.hoverTimer = None
        scene.removeItem(self.hoverText)
        self.hoverText = None

    #def hoverMoveEvent(self, event):
    #    pass

    def mouseMoveEvent(self, event):
        #print "mouseMoveEvent", event
        pos = event.scenePos()
        lastPos = event.lastScenePos()

        self.setX( self.x() - (lastPos.x() - pos.x()) )
        self.setY( self.y() - (lastPos.y() - pos.y()) )

    def mousePressEvent(self, event):
        #print("Node.mousePressEvent()")
        #print event, event.button()
        btn = event.button()
        if btn == Qt.MouseButton.RightButton:
            #event.ignore()
            #print 'trying to ignore...'
            self.popupContextMenu(event)

        elif btn == Qt.MouseButton.LeftButton:
            self.mainW.editor.setFocusedNode(self)




    def popupContextMenu(self, event):
        #print 'Circle.popupContextMenu', event
        menu = QMenu(view)
        a = menu.addAction("Edge...")
        a.triggered.connect(self.drawNewEdge)

        a = menu.addAction("Clone...")
        a.triggered.connect(lambda: self.cloneNode(event))

        menu.addSeparator()
        a = menu.addAction("Delete...")
        a.triggered.connect(self.deleteNode)

        #menu.addSeparator()
        #a = menu.addAction("Find perpendicular to edges[0]...")
        #a.triggered.connect(self.findPerp)

        pos = view.mapToGlobal(view.mapFromScene(event.scenePos()))
        menu.exec_(pos)

    def findPerp(self):
        e = list(edges)[0]
        e.findPerpendicularFromCircle(self)


    def cloneNode(self, event):
        scene.newNode(None, json.loads(self.json["config"]))


    def deleteNode(self):
        self.mainW.deleteNode(self)


    def drawNewEdge(self):
        e = Edge(self.mainW, self, None)
        scene.drawingEdge = e # tell the scene move the tail handle around as it receives mouse events.
        scene.addItem(e)


    def updateCenterCoords(self):
        rect = self.boundingRect()
        self.radius = r = rect.width() / 2
        self.centerX = self.x() + r
        self.centerY = self.y() + rect.height() / 2

        #drawX(self.centerX, self.centerY)


class Handle(QGraphicsSvgItem):
    mainW = None

    def __repr__(self):
        return "<Handle from %r to %r>" % (self.fromNode, self.toNode)

    handlePath = os.path.join(os.path.dirname(__file__), 'svg', 'handle.svg')
    arrowPath = os.path.join(os.path.dirname(__file__), 'svg', 'arrowhead.svg')
    arrowBaseRotation = -45

    # handle at other end of Edge
    otherHandle = None

    def __init__(self, mainW, edge, fromNode, toNode, head=True, *args, **kw):
        QGraphicsSvgItem.__init__(self, self.handlePath if head else self.arrowPath, *args, **kw)
        self.mainW = mainW

        self.edge = edge
        self.head = head

        self.fromNode = fromNode
        self.toNode = toNode

        if fromNode is not None:
            self.fromNode.xChanged.connect(self.nodeMoved)
            self.fromNode.yChanged.connect(self.nodeMoved)

        if toNode is not None:
            self.toNode.xChanged.connect(self.nodeMoved)
            self.toNode.yChanged.connect(self.nodeMoved)

        self.xChanged.connect(self.moved)
        self.yChanged.connect(self.moved)

    def mouseMoveEvent(self, event):
        #print "Handle.mouseMoveEvent()"
        pos = event.scenePos()
        lastPos = event.lastScenePos()

        self.setX( self.x() - (lastPos.x() - pos.x()) )
        self.setY( self.y() - (lastPos.y() - pos.y()) )

        self.otherHandle.nodeMoved()

    def mousePressEvent(self, event):
        #print "Handle.mousePressEvent()"
        #print event, event.button()
        if event.button() == Qt.MouseButton.RightButton:
            #event.ignore()
            #print 'trying to ignore...'
            self.popupContextMenu(event)

    def popupContextMenu(self, event):
        #print 'Handle.popupContextMenu', event
        menu = QMenu(view)
        a = menu.addAction("Delete...")
        a.triggered.connect(self.deleteEdge)

        menu.addSeparator()
        a = menu.addAction("Draw colliding circles...")
        a.triggered.connect(self.drawCollisionsAndTangents)

        pos = view.mapToGlobal(view.mapFromScene(event.scenePos()))
        menu.exec_(pos)

    def findCollisions(self):
        circs = self.edge.findCollidingCircles()
        print('colliding:', [c.nodeID for c in circs])
        return circs

    def drawCollisionsAndTangents(self):
        circs = self.findCollisions()
        for c in circs:
            x,y,x2,y2 = findTangentPointsToCircle(self.originX, self.originY, c.centerX, c.centerY, c.radius)

            drawX(self.originX, self.originY)
            #drawX(cx, cy)
            #drawX(x,y)
            #drawX(x2,y2)


    def deleteEdge(self):
        self.mainW.deleteEdge(self.edge.headHandle.fromNode.nodeID, self.edge.tailHandle.fromNode.nodeID)


    def mouseReleaseEvent(self, event):
        #print "mouseReleaseEvent"
        # Dragged on to another node?
        #print "Handle.mouseReleaseEvent() %r" % (self,)
        
        self.tryDrop(event)

    def tryDrop(self, event):
        """
        The Handle has been moved, and the mouse button release or clicked indicating we want to try and
        "drop" the handle on to a new Node, thus changing the graph edges.

        Return True if we did make a change to the edges, non-True otherwise.
        """
        pos = event.scenePos()

        #print 'pos (%f,%f)' % (pos.x(), pos.y())

        on = []
        for c in self.mainW.nodes.values():
            rect = c.sceneBoundingRect() 
            #print 'rect.center() (%f,%f)' % (rect.center().x(), rect.center().y())
            if distance(pos, rect.center()) < rect.width() / 2:
                print("DROPPED ON %r" % (c,))
                on.append(c)

        if len(on) > 1:
            print("Ambiguous drag/drop. Inside more than one circle. Ignored.")
        elif len(on) == 1:
            c = on[0]
            if c is self.fromNode or c is self.toNode or c is self.otherHandle.toNode:
                print("Can't drop on %r" % (c,))
                self.nodeMoved() # snap handle back to node it is bound to
                self.otherHandle.nodeMoved()
                self.moved()
                return

            # Don't allow drop if there is already an edge between the drop target and the other end of this edge.
            #
            newEdge = [c, self.otherHandle.fromNode]
            newEdge.sort()
            for edge in edges:
                t = [edge.headHandle.fromNode, edge.tailHandle.fromNode]
                t.sort()
                #print(newEdge, t)
                if newEdge == t:
                    self.nodeMoved() # snap handle back to node it is bound to
                    self.otherHandle.nodeMoved()
                    self.moved()
                    print("Edge already exists between %r" % (newEdge,))
                    return

            # OK - make edge change

            self.changeFromNode(c)
            self.otherHandle.changeToNode(c)
            return True

        else:
            self.nodeMoved() # snap handle back to node it is bound to
            self.otherHandle.nodeMoved()
            self.moved()
        

    def changeFromNode(self, n):
        #print('changeFromNode', n)
        if self.fromNode is not None:
            self.fromNode.xChanged.disconnect(self.nodeMoved)
            self.fromNode.yChanged.disconnect(self.nodeMoved)

            if self.head and self.toNode is not None:
                # update JSON document edges.
                self.mainW.deleteEdge( self.fromNode.nodeID, self.toNode.nodeID )

        if self.head and self.toNode is not None:
            # update JSON document edges.
            self.mainW.addEdge( n.nodeID, self.toNode.nodeID, updateGUI=False)

        self.fromNode = n
        self.fromNode.xChanged.connect(self.nodeMoved)
        self.fromNode.yChanged.connect(self.nodeMoved)

        self.nodeMoved()
        self.otherHandle.nodeMoved()


    def changeToNode(self, n):
        #print('changeToNode', n)
        if self.toNode is not None:
            self.toNode.xChanged.disconnect(self.nodeMoved)
            self.toNode.yChanged.disconnect(self.nodeMoved)

            if self.head and self.fromNode is not None:
                # update JSON document edges.
                self.mainW.deleteEdge( self.fromNode.nodeID, self.toNode.nodeID )

        if self.head and self.fromNode is not None:
            # update JSON document edges.
            self.mainW.addEdge( self.fromNode.nodeID, n.nodeID, updateGUI=False )

        self.toNode = n
        self.toNode.xChanged.connect(self.nodeMoved)
        self.toNode.yChanged.connect(self.nodeMoved)

        self.nodeMoved()
        self.otherHandle.nodeMoved()


    def nodeMoved(self):
        """
        We're a handle between two Nodes. One of them moved - update our position
        along the circumfrence of the Node's circle we're attached to.
        """
        #print "Handle.nodeMoved() %r" % (self,)

        if self.fromNode is None:
            return

        r1 = self.fromNode.boundingRect().width() / 2
        #r2 = self.toNode.boundingRect().width() / 2

        c1Center = QPointF(self.fromNode.x() + r1, self.fromNode.y() + r1)
        
        #j = self.toNode.x() + r2
        #k = self.toNode.y() + r2
        j = self.otherHandle.x()
        k = self.otherHandle.y()
        #print '(j,k) == (%r,%r)' % (j,k)

        m = c1Center.x()
        n = c1Center.y()
        #print '(m,n) == (%r,%r)' % (m,n)
        
        M = abs(j-m)
        N = abs(k-n)

        #print('M', repr(M))
        #print('N', repr(N))

        
        z = atan(M/N)

        l = (pi/2) - z

        #print 'l', u'°%r' % (l / (pi/180),)

        Y = r1 * sin(l)
        X = r1 * cos(l)

        if m > j:
            X *= -1
        if n > k:
            Y *= -1

        #print 'here'
        self.originX = m + X
        self.originY = n + Y 
        if self.head:
            self.setX( self.originX - (self.boundingRect().width() / 2))
            self.setY( self.originY - (self.boundingRect().height() / 2))
        else:
            self.setX( self.originX )
            self.setY( self.originY )


    def moved(self):
        """
        We moved for some reason. Update arrowhead angel if we are a tail Handle.
        """
        #print "Handle.moved()"

        if not self.head:
            # we're an arrow handle, not a node handle.
            # so we have to apply a proper rotation.

            if self.otherHandle.head:
                j = self.otherHandle.x() + (self.otherHandle.boundingRect().width() / 2)
                k = self.otherHandle.y() + (self.otherHandle.boundingRect().height() / 2)
            else:
                j = self.otherHandle.x()
                k = self.otherHandle.y()
            #print '(j,k) == (%r,%r)' % (j,k)

            m = self.x()
            n = self.y()
            #print '(m,n) == (%r,%r)' % (m,n)
            
            M = abs(j-m)
            N = abs(k-n)

            if N == 0:
                return

            z = atan(M/N)

            l = (pi/2) - z
            A = l * (180/pi)

            #print 'l', u'°%r' % (l / (pi/180),)

            if m < j and n < k:
                pass
            elif m > j and n < k:
                A = 90 + (90 - A)
            elif m > j and n > k:
                A += 180
            elif m < j and n > k:
                A = 270 + (90 - A)

            self.setRotation( self.arrowBaseRotation + A )


def distance(pos1, pos2):
    return sqrt( (abs(pos2.x() - pos1.x()) ** 2) + (abs(pos2.y() - pos1.y()) ** 2) )

def distanceXY(x, y, x2, y2):
    return sqrt( (x2 - x) ** 2 + (y2 - y) ** 2 )


def findTangentPointsToCircle(x, y, cx, cy, r):
    C = r
    A = sqrt( (cx - x) ** 2 + (cy - y) ** 2 )

    b = acos(C/A)

    X = cos(b) * C
    Y = sin(b) * C

    if cx > x:
        X *= -1
    if cy > y:
        Y *= -1

    return (cx+X, cy+Y)

def findTangentPointsToCircle(x, y, cx, cy, r):
    #find the direction from the point to the center of the circle
    #dir = point - circle_center
    A = sqrt( (cx - x) ** 2 + (cy - y) ** 2 )

    #extract the length and angle
    #len = dir.length
    len = A
    angle = atan2(cy-y, cx-x)
    
    # derive the length of the tangent using pythagoras
    tangent_len = sqrt(len**2 - r**2)
    # and the angle using trigonometry
    tangent_angle = asin(r/len)
    
    # there are 2 tangents, one either side
    pos = angle + tangent_angle
    neg = angle - tangent_angle
            
    #return the direction vector of each tanget (the starting point was passed in)
    return cos(pos), sin(pos), cos(neg), sin(neg)

def findTangentPointsToCircle(x, y, cx, cy, r):
    #print 'x,y   == %.2f,%.2f' % (x,y)
    #print 'cx,cy == %.2f,%.2f' % (cx,cy)
    #print 'r == %.2f' % (r,)
    # The radius of the circle
    #var r1:Number = 30;

    # The center point of the circle
    #var cp:Number = Math.random() * 300+100;
    #var c:Point = new Point(cp, cp);
    # c = (cx, cy)

    # draw circle
    #var mc=new Sprite();
    #mc.graphics.lineStyle(0,0);
    #mc.graphics.drawCircle(0,0,r);
    #mc.x = mc.y = cp;
    #addChild(mc);

    # The point
    #var p = new Point(Math.random() * 500, Math.random() * 400);
    # p = (x,y)

    # Calculate points for intesecting circle
    #c2 = mid point between `c' and `p'
    #var c2:Point = Point.interpolate(c, p, 0.5);
    # c2 = (c2x, c2y)

    if cx > x:
        c2x = x + (cx - x) / 2
    else:
        c2x = cx + (x - cx) / 2

    if cy > y:
        c2y = y + (cy - y) / 2
    else:
        c2y = cy + (y - cy) / 2


    #var r2:Number = Point.distance(c2, c);
    r2 = distanceXY(c2x, c2y, cx, cy)

    #var d:Number = Point.distance(c, c2);
    d = distanceXY(cx, cy, c2x, c2y)

    # Remove comment below to see intersecting circle
    #graphics.beginFill(0xFF0000, 0.25);
    #graphics.drawCircle(c2.x, c2.y, r2);

    #var a:Number = (r*r - r2*r2 + d*d) / (2*d);
    a = (r*r - r2*r2 + d*d) / (2*d);

    #var p2x:Number = c.x + a * ( c2.x - c.x ) / d;
    p2x = cx + a * ( c2x - cx ) / d;
    #var p2y:Number = c.y + a * ( c2.y - c.y ) / d;
    p2y = cy + a * ( c2y - cy ) / d;
    #var h:Number = Math.sqrt(r*r - a*a);
    ok = r*r - a*a
    #print 'ok', ok
    if ok < 0:
        return
    h = sqrt( ok );

    #var d1x:Number = p2x + h * ( c2.y - c.y ) / d;
    d1x = p2x + h * ( c2y - cy ) / d;
    #var d1y:Number = p2y - h * ( c2.x - c.x ) / d;
    d1y = p2y - h * ( c2x - cx ) / d;
    #var d2x:Number = p2x - h * ( c2.y - c.y ) / d;
    d2x = p2x - h * ( c2y - cy ) / d;
    #var d2y:Number = p2y + h * ( c2.x - c.x ) / d;
    d2y = p2y + h * ( c2x - cx ) / d;

    return d1x, d1y, d2x, d2y

    # Draw lines
    #graphics.lineStyle(1, 0xFF00FF, 0.5);
    #graphics.moveTo(p.x, p.y);
    #graphics.lineTo(d1x, d1y);
    #graphics.moveTo(p.x, p.y);
    #graphics.lineTo(d2x, d2y);


def drawX(x, y):
    line = QGraphicsLineItem()
    line.setLine(x-5, y-5, x+5, y+5)
    scene.addItem(line)

    line = QGraphicsLineItem()
    line.setLine(x-5, y+5, x+5, y-5)
    scene.addItem(line)


#def layoutFrom(circle, depth=0):
#    """
#    Really crappy graph layout algorithm.

#    ** Deprecated **
#    """
#    ySpacing = 140
#    xSpacing = 240
#    startY = circle.y() - ( (len(circle.connectedTo)-1) * ySpacing / 2)

#    for i, cNum in enumerate(circle.connectedTo):
#        c = circles[cNum]
#        #print 'i,c', i, c
#        c.setX(depth*xSpacing)
#        c.setY(startY + i * ySpacing )
#        #print 'done'

#        layoutFrom(c, depth=depth+1)


def getControlPoints(x0, y0, x1, y1, x2, y2, t):
    #  x0,y0,x1,y1 are the coordinates of the end (knot) pts of this segment
    #  x2,y2 is the next knot -- not connected here but needed to calculate p2
    #  p1 is the control point calculated here, from x1 back toward x0.
    #  p2 is the next control point, calculated here and returned to become the 
    #  next segment's p1.
    #  t is the 'tension' which controls how far the control points spread.
    
    # Scaling factors: distances from this knot to the previous and following knots.
    d01 = sqrt(pow(x1-x0, 2) + pow(y1-y0, 2))
    d12 = sqrt(pow(x2-x1, 2) + pow(y2-y1, 2))
   
    fa = t * d01 / (d01 + d12)
    fb = t - fa
  
    p1x = x1 + fa * (x0 - x2)
    p1y = y1 + fa * (y0 - y2)

    p2x = x1 - fb * (x0 - x2)
    p2y = y1 - fb * (y0 - y2)
    
    return [p1x, p1y, p2x, p2y]


def drawSpline(qppath, pts, t):
    cp=[]   # array of control points, as x0,y0,x1,y1,...
    n = len(pts)

    i = 0
    while i < n - 4:
        cp += getControlPoints(pts[i], pts[i+1], pts[i+2], pts[i+3], pts[i+4], pts[i+5], t)
        i += 2

    i = 2
    while i < len(pts) - 5:
        qppath.moveTo(pts[i], pts[i+1])
        qppath.cubicTo(cp[2*i-2], cp[2*i-1], cp[2*i], cp[2*i+1], pts[i+2], pts[i+3])

        i += 2

    qppath.moveTo(pts[0], pts[1])
    qppath.quadTo(cp[0], cp[1], pts[2], pts[3]);

    qppath.moveTo(pts[n-2], pts[n-1])
    qppath.quadTo(cp[2*n-10], cp[2*n-9], pts[n-4], pts[n-3]);


scene = None
view = None
edges = set()


def usage():
    print("Usage: catgut [ <graph-file> ]", file=sys.stderr)
    

def main():
    app = QtGui.QApplication(sys.argv)

    mainW = CatgutMainWindow(app)

    global scene
    scene = CatgutGraphicsScene(mainW)
    global view
    view = QGraphicsView(scene)

    mainW.setCentralWidget(view)

    sidebar = QSplitter()
    sidebar.setOrientation(Qt.Vertical)

    editor = CatgutNodeEditor(mainW)
    mainW.editor = editor


    sidebar.addWidget(editor)

    colorkey = CatgutColorkeyWidget()

    sidebar.addWidget(colorkey)


    dockable = QDockWidget("Node Editor", parent=mainW)
    dockable.setObjectName("Node Editor")
    dockable.setFeatures(QDockWidget.DockWidgetMovable | QDockWidget.DockWidgetFloatable)
    dockable.setWidget(sidebar)

    mainW.addDockWidget(Qt.LeftDockWidgetArea, dockable)


    if len(sys.argv) > 2:
        usage()
        sys.exit(1)

    if len(sys.argv) == 2:
        if sys.argv[1] in ["-v", "--version"]:
            print(VERSION)
            sys.exit(0)
        
        mainW.filename = sys.argv[1]
        mainW.doOpenFile()
    else:
        mainW.newGraph()


    # load persisted main window geometry and dockable widgets' positions
    mainW.readSettings()

    #TODO do something that actually works to size the default window
    # to fit the extends of the displayed graph
    #
    #r = scene.itemsBoundingRect()
    #print('r', r)
    #view.setSceneRect( r )
    #size = mainW.size()
    #width = size.width()
    #width -= view.size().width()
    #width += scene

    global edges

    mainW.show()

    sizes = sidebar.sizes()
    total = sum(sizes)
    sizes[0] = 1
    sizes[-1] = total - sizes[0]
    sidebar.setSizes(sizes)

    mainW.setWindowTitle("CATGUT - the CATtura Graph Unfucking Tool")

    iconPath = os.path.join(os.path.dirname(__file__), 'catgut.png')
    mainW.setWindowIcon(QIcon(iconPath))
    sys.exit(app.exec_())
